(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2013-2016)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)
theory Critical_Pair_Closure_Impl
imports
  Critical_Pairs_Impl
  Parallel_Closed
  Strongly_Closed
  "../Framework/QDP_Framework_Impl"
begin

definition check_strongly_closed where
  "check_strongly_closed R n = do {
     check_linear_trs R;
     check_allm (\<lambda> (b, s, t). do {
       check (\<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R s n) x) (reachable_terms R t (Suc 0))) \<and>
         \<not> Option.is_none (List.find (\<lambda>x. List.member (reachable_terms R t n) x) (reachable_terms R s (Suc 0))))
       (shows ''the critical pair '' +@+ shows_term s +@+ shows '' <- . -> '' +@+ shows_term t +@+
        shows '' is not strongly closed within '' +@+ shows n +@+ shows '' steps.'')
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s +@+ shows_nl +@+ shows ''hence the following TRS is not strongly closed '' +@+
       shows_nl +@+ shows_trs R +@+ indent s)"

lemma check_strongly_closed:
  assumes "isOK(check_strongly_closed R n)"
  shows "CR (rstep (set R))"
proof (rule strongly_closed_linear_CR)
  from assms show "linear_trs (set R)" unfolding check_strongly_closed_def by auto
  let ?red = "\<lambda>s n. reachable_terms R s n"
  let ?opt = "\<lambda>s t. List.find (\<lambda>x. List.member (?red s n) x) (?red t (Suc 0))"
  { fix s t
    assume "\<not> Option.is_none (?opt s t)"
    then obtain r where r: "?opt s t = Some r" by (cases "?opt s t") auto
    then have "List.member (?red s n) r" unfolding find_Some_iff by auto
    then have mem:"r \<in> set (?red s n)" using in_set_member by fast
    from r have "\<exists>i<length (?red t (Suc 0)). r = (?red t (Suc 0)) ! i" unfolding find_Some_iff by auto
    then have "r \<in> set (?red t (Suc 0))" using nth_mem by blast
    with mem have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      using reachable_terms reachable_terms_one by metis
  } note sound=this
  { fix b s t
    assume "(b, s, t) \<in> critical_pairs (set R) (set R)"
    with assms have *:"\<not> Option.is_none (?opt s t) \<and> \<not> Option.is_none (?opt t s)"
      unfolding check_strongly_closed_def by fastforce
    with sound have "\<exists>r. (s, r) \<in> (rstep (set R))\<^sup>* \<and> (t, r) \<in> (rstep (set R))\<^sup>="
      and  "\<exists>u. (t, u) \<in> (rstep (set R))\<^sup>* \<and> (s, u) \<in> (rstep (set R))\<^sup>=" by blast+
  }
  then show "strongly_closed (set R)" unfolding strongly_closed_def by blast
qed

definition is_root_step :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where 
  "is_root_step R s t = (\<exists> (l, r) \<in> set R. case match s l of
      None \<Rightarrow> False
    | Some \<sigma> \<Rightarrow> r \<cdot> \<sigma> = t)"

lemma is_root_step:
  assumes "is_root_step R s t"
  shows "(s, t) \<in> rrstep (set R)"
using assms
unfolding is_root_step_def rrstep_def rstep_r_p_s_def' 
by (auto split: option.splits) (force dest: match_matches)

fun is_par_rstep :: "('f, 'v) rules \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "is_par_rstep R (Var x) (Var y) = (x = y)"
| "is_par_rstep R (Fun f ss) (Fun g ts) = 
    (Fun f ss = Fun g ts \<or> is_root_step R (Fun f ss) (Fun g ts) \<or>
    (if f = g \<and> length ss = length ts then list_all2 (is_par_rstep R) ss ts else False))"
| "is_par_rstep R s t = is_root_step R s t"

lemma is_par_rstep:
  assumes "is_par_rstep R s t"
  shows "(s, t) \<in> par_rstep (set R)"
using assms
proof (induct rule: is_par_rstep.induct)
  case (1 R x y) then show ?case by auto
next
  case (2 R f ss g ts)
  then consider (eq) "Fun f ss = Fun g ts" 
    | (root) "is_root_step R (Fun f ss) (Fun g ts)"
    | (args) "f = g" and "length ss = length ts" and "list_all2 (is_par_rstep R) ss ts"
    by (auto split: if_splits)
  then show ?case
  proof (cases)
    case eq then show ?thesis using par_rstep_refl by auto
  next
    case root 
    from is_root_step[OF this] obtain l r \<sigma> 
      where "Fun f ss = l \<cdot> \<sigma>" and "Fun g ts = r \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
    then show ?thesis using par_rstep.root_step[OF lr] by metis
  next
    case args
    { fix i
      assume i:"i < length ss"
      then have si:"ss ! i \<in> set ss" by auto
      from i args(2) have ti:"ts ! i \<in> set ts" by auto
      from args(3) have "is_par_rstep R (ss ! i) (ts ! i)" using i list_all2_nthD by blast
      with 2(1)[of "ss ! i" "ts ! i"] args(1,2) si ti have "(ss ! i, ts ! i) \<in> par_rstep (set R)"
        by auto
    } 
    then show ?thesis using args(1,2) by (auto)
  qed
next
  case ("3_1" R f ss y)
  then have "is_root_step R (Fun f ss) (Var y)" by auto
  from is_root_step[OF this] obtain l r \<sigma> 
      where "Fun f ss = l \<cdot> \<sigma>" and "Var y = r \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
  then show ?case using par_rstep.root_step[OF lr] by metis
next
  case ("3_2" R y f ss)
  then have "is_root_step R (Var y) (Fun f ss)" by auto
  from is_root_step[OF this] obtain l r \<sigma> 
      where "Fun f ss = r \<cdot> \<sigma>" and "Var y = l \<cdot> \<sigma>" and lr:"(l, r) \<in> set R"
      unfolding rrstep_def' by auto
  then show ?case using par_rstep.root_step[OF lr] by metis
qed

fun is_parallel_closed_cp where
  "is_parallel_closed_cp R _ (False, s, t) = is_par_rstep R t s"
| "is_parallel_closed_cp R n (True, s, t) = 
    (\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R t v) (reachable_terms R s n)))"

definition check_parallel_closed where
  "check_parallel_closed R n = do {
     check_left_linear_trs R;
     check_allm (\<lambda> (l,r). check (is_Fun l) (shows ''the TRS has variables as left-hand sides'')) R;
     check_allm (\<lambda> (b, s, t). do {
       check (is_parallel_closed_cp R n (b, s, t))
         (shows ''the critical pair '' +@+ shows_term s +@+ shows '' <- . -> '' +@+ shows_term t +@+
          shows '' is not (almost) parallel closed within '' +@+ shows n +@+ shows '' steps.'')
     }) (critical_pairs_impl R R)
     }  <+? (\<lambda>s. s +@+ shows_nl +@+ shows ''hence the following TRS is not (almost) parallel closed '' +@+
       shows_nl +@+ shows_trs R +@+ indent s)"

lemma check_parallel_closed:
  assumes "isOK(check_parallel_closed R n)"
  shows "CR (rstep (set R))"
proof (rule parallel_closed_imp_CR)
  from assms show "left_linear_trs (set R)" unfolding check_parallel_closed_def by auto
  from assms show "\<And>l r. (l, r) \<in> set R \<Longrightarrow> is_Fun l" unfolding check_parallel_closed_def by auto
  show "\<And>p q. (False, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow> (q, p) \<in> par_rstep (set R)"
  proof -
    fix p q
    assume "(False, p, q) \<in> critical_pairs (set R) (set R)"
    with assms have "is_par_rstep R q p" unfolding check_parallel_closed_def by auto
    then show "(q, p) \<in> par_rstep (set R)" using is_par_rstep by auto
  qed
  show "\<And>p q. (True, p, q) \<in> critical_pairs (set R) (set R) \<Longrightarrow>
           \<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
  proof -
    fix p q
    assume "(True, p, q) \<in> critical_pairs (set R) (set R)"
    with assms have "\<not> Option.is_none (List.find (\<lambda>v. is_par_rstep R q v) (reachable_terms R p n))"
      unfolding check_parallel_closed_def by auto
    then obtain v where  "List.find (\<lambda>v. is_par_rstep R q v) (reachable_terms R p n) = Some v"
      by force
    then have "is_par_rstep R q v" and "v \<in> set (reachable_terms R p n)"
      unfolding find_Some_iff by auto
    then show "\<exists>v. (q, v) \<in> par_rstep (set R) \<and> (p, v) \<in> (rstep (set R))\<^sup>*"
      using is_par_rstep reachable_terms by blast
   qed
qed

end
